@@ -20,7 +20,7 @@
 ///         Pre-sieving provides a speedup of up to 30% when
 ///         sieving the primes < 10^10 using primesieve.
 ///
-/// Copyright (C) 2022 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2023 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
@@ -33,7 +33,21 @@
 #include <stdint.h>
 #include <algorithm>
 #include <cmath>
-#include <vector>
+#include <initializer_list>
+
+/// All x64 CPUs support the SSE2 vector instruction set
+#if defined(__SSE2__) && \
+    __has_include(<emmintrin.h>)
+  #include <emmintrin.h>
+  #define HAS_SSE2
+#endif
+
+// All ARM64 CPUs support the NEON vector instruction set
+#if (defined(__ARM_NEON) || defined(__aarch64__)) && \
+     __has_include(<arm_neon.h>)
+  #include <arm_neon.h>
+  #define HAS_ARM_NEON
+#endif
 
 using std::copy_n;
 using primesieve::pod_array;
@@ -147,7 +161,7 @@ const pod_array<uint8_t, 7*11*13> buffer_7_11_13 =
 };
 
 /// Pre-sieve with the primes < 100
-const pod_array<std::vector<uint64_t>, 8> bufferPrimes =
+const pod_array<std::initializer_list<uint64_t>, 8> bufferPrimes =
 {{
   {  7, 67, 71 },  // 32 KiB
   { 11, 41, 73 },  // 32 KiB
@@ -170,23 +184,116 @@ const uint64_t buffersDist =
        (79 * 97) * 30 +
        (83 * 89) * 30;
 
-void andBuffers(const uint8_t* __restrict buf1,
+#if defined(HAS_SSE2)
+
+/// Since compiler auto-vectorization is not 100% reliable, we have
+/// manually vectorized the andBuffers() function for x64 CPUs.
+/// This algorithm is portable since all x64 CPUs support the SSE2
+/// instruction set.
+///
+void andBuffers(const uint8_t* __restrict buf0,
+                const uint8_t* __restrict buf1,
                 const uint8_t* __restrict buf2,
                 const uint8_t* __restrict buf3,
                 const uint8_t* __restrict buf4,
                 const uint8_t* __restrict buf5,
                 const uint8_t* __restrict buf6,
                 const uint8_t* __restrict buf7,
-                const uint8_t* __restrict buf8,
                 uint8_t* __restrict output,
                 std::size_t bytes)
 {
-  // This loop should be auto-vectorized
-  for (std::size_t i = 0; i < bytes; i++)
-    output[i] = buf1[i] & buf2[i] & buf3[i] & buf4[i]
-              & buf5[i] & buf6[i] & buf7[i] & buf8[i];
+  std::size_t i = 0;
+  std::size_t limit = bytes - bytes % sizeof(__m128i);
+
+  // Note that I also tried vectorizing this algorithm using AVX2
+  // which has double the vector width compared to SSE2, but this did
+  // not provide any speedup. On average, this loop processes only
+  // 2192 bytes, hence there aren't many vector loop iterations and
+  // by increasing the vector width this also increases the number of
+  // scalar loop iterations after the vector loop finishes which
+  // could potentially even become a bottleneck.
+  for (; i < limit; i += sizeof(__m128i))
+  {
+    _mm_storeu_si128((__m128i*) &output[i],
+        _mm_and_si128(
+            _mm_and_si128(
+                _mm_and_si128(_mm_loadu_si128((const __m128i*) &buf0[i]), _mm_loadu_si128((const __m128i*) &buf1[i])),
+                _mm_and_si128(_mm_loadu_si128((const __m128i*) &buf2[i]), _mm_loadu_si128((const __m128i*) &buf3[i]))),
+            _mm_and_si128(
+                _mm_and_si128(_mm_loadu_si128((const __m128i*) &buf4[i]), _mm_loadu_si128((const __m128i*) &buf5[i])),
+                _mm_and_si128(_mm_loadu_si128((const __m128i*) &buf6[i]), _mm_loadu_si128((const __m128i*) &buf7[i])))));
+  }
+
+  for (; i < bytes; i++)
+    output[i] = buf0[i] & buf1[i] & buf2[i] & buf3[i] &
+                buf4[i] & buf5[i] & buf6[i] & buf7[i];
 }
 
+#elif defined(HAS_ARM_NEON)
+
+/// Homebrew compiles its C/C++ packages on macOS using Clang -Os
+/// (instead of -O2 or -O3) which does not auto-vectorize our simple
+/// loop with Bitwise AND. If this loop is not vectorized this
+/// deteriorates the performance of primesieve by up to 40%. As a
+/// workaround for this Homebrew issue we have manually vectorized
+/// the Bitwise AND loop using ARM NEON.
+///
+void andBuffers(const uint8_t* __restrict buf0,
+                const uint8_t* __restrict buf1,
+                const uint8_t* __restrict buf2,
+                const uint8_t* __restrict buf3,
+                const uint8_t* __restrict buf4,
+                const uint8_t* __restrict buf5,
+                const uint8_t* __restrict buf6,
+                const uint8_t* __restrict buf7,
+                uint8_t* __restrict output,
+                std::size_t bytes)
+{
+  std::size_t i = 0;
+  std::size_t limit = bytes - bytes % sizeof(uint8x16_t);
+
+  for (; i < limit; i += sizeof(uint8x16_t))
+  {
+    vst1q_u8(&output[i],
+        vandq_u8(
+            vandq_u8(
+                vandq_u8(vld1q_u8(&buf0[i]), vld1q_u8(&buf1[i])),
+                vandq_u8(vld1q_u8(&buf2[i]), vld1q_u8(&buf3[i]))),
+            vandq_u8(
+                vandq_u8(vld1q_u8(&buf4[i]), vld1q_u8(&buf5[i])),
+                vandq_u8(vld1q_u8(&buf6[i]), vld1q_u8(&buf7[i])))));
+  }
+
+  for (; i < bytes; i++)
+    output[i] = buf0[i] & buf1[i] & buf2[i] & buf3[i] &
+                buf4[i] & buf5[i] & buf6[i] & buf7[i];
+}
+
+#else
+
+void andBuffers(const uint8_t* __restrict buf0,
+                const uint8_t* __restrict buf1,
+                const uint8_t* __restrict buf2,
+                const uint8_t* __restrict buf3,
+                const uint8_t* __restrict buf4,
+                const uint8_t* __restrict buf5,
+                const uint8_t* __restrict buf6,
+                const uint8_t* __restrict buf7,
+                uint8_t* __restrict output,
+                std::size_t bytes)
+{
+  // This loop will get auto-vectorized if compiled with GCC/Clang
+  // using -O3. Using GCC -O2 does not auto-vectorize this loop
+  // because -O2 uses the "very-cheap" vector cost model. To fix
+  // this issue we enable -ftree-vectorize -fvect-cost-model=dynamic
+  // if the compiler supports it in auto_vectorization.cmake.
+  for (std::size_t i = 0; i < bytes; i++)
+    output[i] = buf0[i] & buf1[i] & buf2[i] & buf3[i] &
+                buf4[i] & buf5[i] & buf6[i] & buf7[i];
+}
+
+#endif
+
 } // namespace
 
 namespace primesieve {
@@ -237,7 +344,7 @@ void PreSieve::initBuffers()
     uint64_t stop = start + product;
     buffers_[i].resize(product / 30);
     std::fill(buffers_[i].begin(), buffers_[i].end(), 0xff);
-    uint64_t maxPrime = bufferPrimes[i].back();
+    uint64_t maxPrime = *(bufferPrimes[i].end() - 1);
     ASSERT(maxPrime == *std::max_element(bufferPrimes[i].begin(), bufferPrimes[i].end()));
     ASSERT(start >= maxPrime * maxPrime);
     maxPrime_ = std::max(maxPrime_, maxPrime);
